// written in the D programming language

module didisplay;

import std.stdio;
import std.string;
import std.ascii : isDigit;
import std.uni : toLower;
import std.conv : to;
private import std.utf : count;
private import std.math : floor;

import config;
import options;
import dispopts;
import diskpart;
import dilocale;

private:

enum char
  DI_SORT_NONE = 'n',
  DI_SORT_MOUNT = 'm',
  DI_SORT_SPECIAL = 's',
  DI_SORT_AVAIL = 'a',
  DI_SORT_REVERSE = 'r',
  DI_SORT_TYPE = 't',
  DI_SORT_ASCENDING = 1;

enum dchar
  FMT_MOUNT = 'm',
  FMT_MOUNT_FULL = 'M',
  FMT_SPECIAL = 's',
  FMT_SPECIAL_FULL = 'S',
  FMT_TYPE = 't',
  FMT_TYPE_FULL = 'T',
  FMT_BLOCKS_TOT = 'b',
  FMT_BLOCKS_TOT_AVAIL = 'B',
  FMT_BLOCKS_USED = 'u',
  FMT_BLOCKS_CALC_USED = 'c',
  FMT_BLOCKS_FREE = 'f',
  FMT_BLOCKS_AVAIL = 'v',
  FMT_BLOCKS_PERC_NOT_AVAIL = 'p',
  FMT_BLOCKS_PERC_USED = '1',
  FMT_BLOCKS_PERC_USED_AVAIL = '2',
  FMT_BLOCKS_PERC_AVAIL = 'a',
  FMT_BLOCKS_PERC_FREE = '3',
  FMT_INODES_TOT = 'i',
  FMT_INODES_USED = 'U',
  FMT_INODES_FREE = 'F',
  FMT_INODES_PERC = 'P',
  FMT_MOUNT_TIME = 'I',
  FMT_MOUNT_OPTIONS = 'O';

enum int
  FTYPE_STRING = 1,
  FTYPE_SPACE = 2,
  FTYPE_PERC_SPACE = 3,
  FTYPE_INODE = 4,
  FTYPE_PERC_INODE = 5;

struct FormatInfo {
  dchar         key;
  int           ftype;
  long          width;
  string        title;
};

FormatInfo[] formatTypes = [
  { FMT_MOUNT,                  FTYPE_STRING,    15, "Mount" },
  { FMT_MOUNT_FULL,             FTYPE_STRING,    15, "Mount" },
  { FMT_SPECIAL,                FTYPE_STRING,    18, "Filesystem" },
  { FMT_SPECIAL_FULL,           FTYPE_STRING,    18, "Filesystem" },
  { FMT_TYPE,                   FTYPE_STRING,     7, "fsType" },
  { FMT_TYPE_FULL,              FTYPE_STRING,     7, "fs Type" },
  { FMT_BLOCKS_TOT,             FTYPE_SPACE,      8, "" },
  { FMT_BLOCKS_TOT_AVAIL,       FTYPE_SPACE,      8, "" },
  { FMT_BLOCKS_USED,            FTYPE_SPACE,      8, "Used" },
  { FMT_BLOCKS_CALC_USED,       FTYPE_SPACE,      8, "Used" },
  { FMT_BLOCKS_FREE,            FTYPE_SPACE,      8, "Free" },
  { FMT_BLOCKS_AVAIL,           FTYPE_SPACE,      8, "Avail" },
  { FMT_BLOCKS_PERC_NOT_AVAIL,  FTYPE_PERC_SPACE, 6, "%Used" },
  { FMT_BLOCKS_PERC_USED,       FTYPE_PERC_SPACE, 6, "%Used" },
  { FMT_BLOCKS_PERC_USED_AVAIL, FTYPE_PERC_SPACE, 6, "%Used" },
  { FMT_BLOCKS_PERC_AVAIL,      FTYPE_PERC_SPACE, 5, "%Free" },
  { FMT_BLOCKS_PERC_FREE,       FTYPE_PERC_SPACE, 5, "%Free" },
  { FMT_INODES_TOT,             FTYPE_INODE,      9, "Inodes" },
  { FMT_INODES_USED,            FTYPE_INODE,      9, "IUsed" },
  { FMT_INODES_FREE,            FTYPE_INODE,      9, "IFree" },
  { FMT_INODES_PERC,            FTYPE_PERC_INODE, 6, "%IUsed" },
  { FMT_MOUNT_TIME,             FTYPE_STRING,    15, "Mount Time" },
  { FMT_MOUNT_OPTIONS,          FTYPE_STRING,    15, "Options" }
];

size_t[dchar]   formatTypesIdxs;
bool            fmtInfoIdxsInitialized;

struct DisplayData
{
  Options           opts;
  DisplayOpts       dispOpts;
  DiskPartitions    dpList;
  string            titleString;
  string[dchar]     dispFmtString;
};

struct DisplayTable {
  real              size;
  real              low;
  real              high;
  short             precision;
  char              key;
  string            suffix;
  string            disp[2];
};

DisplayTable[] displayTable = [
  // size == low except for bytes
  // size,low,high,precision,key,suffix,disp[1000,1024]
  { 0, 0, 0, 0, 'b', " ", [ "Bytes", "Bytes" ] },
  { 0, 0, 0, 0, 'k', "K", [ "KBytes", "KBytes" ] },
  { 0, 0, 0, 1, 'm', "M", [ "Megs", "Mebis" ] },
  { 0, 0, 0, 1, 'g', "G", [ "Gigs", "Gibis" ] },
  { 0, 0, 0, 1, 't', "T", [ "Teras", "Tebis" ] },
  { 0, 0, 0, 1, 'p', "P", [ "Petas", "Pebis" ] },
  { 0, 0, 0, 1, 'e', "E", [ "Exas", "Exbis" ] },
  { 0, 0, 0, 1, 'z', "Z", [ "Zettas", "Zebis" ] },
  { 0, 0, 0, 1, 'y', "Y", [ "Yottas", "Yobis" ] },
  { 0, 0, 0, 1, 'h', "h", [ "Size", "Size" ] },
  { 0, 0, 0, 1, 'H', "H", [ "Size", "Size" ] }
  ];

size_t[char]    displayIdxs;
bool            dispIdxsInitialized;

public:

void
doDisplay (Options opts, ref DisplayOpts dispOpts,
    ref DiskPartitions dpList, bool hasPooled)
{
  DisplayData       dispData;

  initializeIdxs ();
  initializeTitles (dispOpts);
  initializeDisplayTable (dispOpts);
  setDispBlockSize (dispOpts);
  if (opts.debugLevel > 30)
  {
    dumpDispTable ();
  }

  dispData.opts = opts;
  dispData.dispOpts = dispOpts;
  dispData.dpList = dpList;

  buildDisplayList (dispData);
  displayTitle (dispData);
  displayPartitions (dispData, hasPooled);
}

void
sortPartitions (ref DiskPartition[] dps,
        ref size_t[] sortIndex, string sortType)
{
  int    gap;

  auto count = dps.length;
  if (count <= 1)
  {
    return;
  }

  gap = 1;
  while (gap < count)
  {
    gap = 3 * gap + 1;
  }

  for (gap /= 3; gap > 0; gap /= 3)
  {
    for (int i = gap; i < count; ++i)
    {
      auto tempIndex = sortIndex[i];
      auto j = i - gap;

      while (j >= 0 &&
            dpCompare (dps, sortType, sortIndex[j], tempIndex) > 0)
      {
        sortIndex[j+gap] = sortIndex[j];
        j = j - gap;
      }

      j += gap;
      if (j != i)
      {
        sortIndex[j] = tempIndex;
      }
    }
  }
}

private:

void
dumpDispTable ()
{
  foreach (disp; displayTable)
  {
    writefln ("%s %s %-6s sz:%25.0f\n  low:%25.0f high:%25.0f",
        disp.key, disp.suffix, disp.disp[0], disp.size, disp.low, disp.high);
  }
}

void
buildDisplayList (ref DisplayData dispData)
{
  int[dchar]    dpMax;
  bool          first = true;

  foreach (dchar c; dispData.opts.formatString)
  {
    if (c in formatTypesIdxs &&
        formatTypes[formatTypesIdxs[c]].ftype == FTYPE_STRING)
    {
      dpMax[c] = 0;
    }
  }

  foreach (dp; dispData.dpList.diskPartitions)
  {
    if (! dp.doPrint) {
      continue;
    }

    if (FMT_SPECIAL_FULL in dpMax) {
      setMaxLen (dp.special, dpMax[FMT_SPECIAL_FULL]);
    }
    if (FMT_MOUNT_FULL in dpMax) {
      setMaxLen (dp.name, dpMax[FMT_MOUNT_FULL]);
    }
    if (FMT_TYPE_FULL in dpMax) {
      setMaxLen (dp.fsType, dpMax[FMT_TYPE_FULL]);
    }
    if (FMT_MOUNT_OPTIONS in dpMax) {
      setMaxLen (dp.mountOptions, dpMax[FMT_MOUNT_OPTIONS]);
    }
    if (FMT_MOUNT_TIME in dpMax) {
      setMaxLen (dp.mountTime, dpMax[FMT_MOUNT_TIME]);
    }
  }

  foreach (dchar c; dispData.opts.formatString)
  {
    if (c in formatTypesIdxs) {
      string fmt;
      auto title = DI_GT (formatTypes[formatTypesIdxs[c]].title);
      auto ftype = formatTypes[formatTypesIdxs[c]].ftype;
      auto width = formatTypes[formatTypesIdxs[c]].width;

      switch (c)
      {
        case FMT_MOUNT_FULL:
        case FMT_SPECIAL_FULL:
        case FMT_TYPE_FULL:
        case FMT_MOUNT_TIME:
        case FMT_MOUNT_OPTIONS:
        {
          width = dpMax[c];
          break;
        }

        case FMT_BLOCKS_TOT:
        case FMT_BLOCKS_TOT_AVAIL:
        {
          title = DI_GT (dispData.dispOpts.dispBlockLabel);
          break;
        }

        default:
        {
          break;
        }
      }

      if (dispData.opts.csvOutput) {
        title = to!string(c);
      }

      auto utflen = std.utf.count (title);
      if (utflen > width)
      {
        width = utflen;
      }

      switch (ftype)
      {
        case FTYPE_STRING:
        {
          fmt = format ("%%-%ds ", width);
          if (dispData.opts.csvOutput) {
            fmt = "\"%s\"";
          }
          dispData.dispFmtString[c] = fmt;
          break;
        }

        case FTYPE_SPACE:
        {
          if (dispData.dispOpts.width != 0) {
            width = dispData.dispOpts.width;
          }
          auto twidth = width;
          if (dispData.dispOpts.dbsstr == "h" || dispData.dispOpts.dbsstr == "H")
          {
            --twidth;
          }
          fmt = format ("%%%d.%df", twidth, dispData.dispOpts.precision);
          if (dispData.opts.csvOutput) {
            fmt = format ("%%.%df", dispData.dispOpts.precision);
          }
          dispData.dispFmtString[c] = fmt;
          break;
        }

        case FTYPE_PERC_SPACE:
        {
          auto twidth = width - 2;
          if (twidth < 1) { twidth = 1; }
          fmt = format ("%%%d.0f%%%%  ", twidth);
          if (dispData.opts.csvOutput) {
            fmt = "%.0f%%";
          }
          dispData.dispFmtString[c] = fmt;
          break;
        }

        case FTYPE_INODE:
        {
          if (dispData.dispOpts.inodeWidth != 0) {
            width = dispData.dispOpts.inodeWidth;
          }
          fmt = format ("%%%d.0f ", width);
          if (dispData.opts.csvOutput) {
            fmt = "%.0f";
          }
          dispData.dispFmtString[c] = fmt;
          break;
        }

        case FTYPE_PERC_INODE:
        {
          fmt = format ("%%%d.0f%%%% ", width);
          if (dispData.opts.csvOutput) {
            fmt = "%.0f%%";
          }
          dispData.dispFmtString[c] = fmt;
          break;
        }

        default:
        {
          break;
        }
      } // switch on format type

      auto twid = width + title.length - utflen;
      if (dispData.opts.csvOutput) {
        fmt = "%s";
      } else {
        if (ftype == FTYPE_STRING) {
          fmt = format ("%%-%ds ", twid);
        } else {
          fmt = format ("%%%ds ", twid);
        }
      }
      if (dispData.opts.csvOutput && ! first) {
        dispData.titleString ~= ",";
      }
      dispData.titleString ~= format (fmt, title);

    } else {
      dispData.titleString ~= format ("%s", c);
    }
    first = false;
  } // for each format string char
}

void
displayTitle (ref DisplayData dispData)
{
  if (! dispData.opts.printHeader) { return; }
  writeln (dispData.titleString);
}


void
displayPartitions (ref DisplayData dispData, bool hasPooled)
{
  DiskPartition     totdp;
  size_t[]          sortIndex;

  sortIndex.length = dispData.dpList.diskPartitions.length;
  for (int i = 0; i < sortIndex.length; ++i) {
    sortIndex[i] = i;
  }

  if (dispData.opts.displayTotal) {
    bool      inpool;
    string    lastpool;

    totdp.initDiskPartition;
    totdp.name = DI_GT("Total");

    if (hasPooled) {
      sortPartitions (dispData.dpList.diskPartitions, sortIndex, "s");
    }

    foreach (size_t i, size_t v; sortIndex)
    {
      bool      startpool;
      auto      dp = dispData.dpList.diskPartitions[v];

      startpool = false;

      if (hasPooled && dp.isPooledFS) {
        if (lastpool.length == 0 ||
            (dp.special.length >= lastpool.length &&
             lastpool != dp.special[0..lastpool.length])) {
          lastpool = dp.special;
          auto si = indexOf (lastpool, "#");  /* for advfs */
          if (si >= 0) {
            lastpool.length = si;
          }
          si = indexOf (lastpool, "/");  /* for zfs */
          if (si >= 0) {
            lastpool.length = si;
          }
          inpool = false;
          startpool = true;
        }
      } else {
        inpool = false;
      }

      if (dp.doPrint) {
        /* only add in a pooled filesystem if it's the first in the list    */
        /* belonging to that pool                                           */
        if (! inpool) {
          totdp.totalBlocks += dp.totalBlocks;
          totdp.freeBlocks += dp.freeBlocks;
          totdp.availBlocks += dp.availBlocks;
          totdp.totalInodes += dp.totalInodes;
          totdp.freeInodes += dp.freeInodes;
          totdp.availInodes += dp.availInodes;
        } else {
          /* if in a pool of disks, add the total used to the totals */
          totdp.totalBlocks += dp.totalBlocks - dp.freeBlocks;
          totdp.totalInodes += dp.totalInodes - dp.freeInodes;
        }
      }

      if (startpool) {
        inpool = true;
      }
    }
  }

  sortPartitions (dispData.dpList.diskPartitions, sortIndex,
        dispData.opts.sortType);

  foreach (size_t i, size_t v; sortIndex)
  {
    auto dp = dispData.dpList.diskPartitions[v];

    if (! dp.doPrint) {
      continue;
    }

    printPartition (dispData, dp);
  } // for each disk partition

  if (dispData.opts.displayTotal) {
    printPartition (dispData, totdp);
  }
}

void
printPartition (ref DisplayData dispData, DiskPartition dp)
{
  string      outString;
  string      sval;
  real        rval;
  real        uval;
  size_t      dHidx;
  bool        first = true;

  if (dispData.dispOpts.dbsstr == "H") {
    real mval = 0.0;

    foreach (dchar c; dispData.opts.formatString)
    {
      if (c in formatTypesIdxs) {
        auto ftype = formatTypes[formatTypesIdxs[c]].ftype;

        if (ftype == FTYPE_SPACE) {
          switch (c)
          {
            case FMT_BLOCKS_TOT:
            {
              rval = dp.totalBlocks;
              break;
            }

            case FMT_BLOCKS_TOT_AVAIL:
            {
              rval = dp.totalBlocks - (dp.freeBlocks - dp.availBlocks);
              break;
            }

            case FMT_BLOCKS_USED:
            {
              rval = dp.totalBlocks - dp.freeBlocks;
              break;
            }

            case FMT_BLOCKS_CALC_USED:
            {
              rval = dp.totalBlocks - dp.availBlocks;
              break;
            }

            case FMT_BLOCKS_FREE:
            {
              rval = dp.freeBlocks;
              break;
            }

            case FMT_BLOCKS_AVAIL:
            {
              rval = dp.availBlocks;
              break;
            }

            default:
            {
              break;
            }
          }  // switch on format character

          if (rval > mval) {
            mval = rval;
          }
        }  // is FTYPE_SPACE
      }  // valid format character
    } // for each character in format string

    dHidx = findDispSize (mval);
  } // if -d H

  foreach (dchar c; dispData.opts.formatString)
  {
    if (c in formatTypesIdxs) {
      auto ftype = formatTypes[formatTypesIdxs[c]].ftype;

      switch (c)
      {
        case FMT_MOUNT:
        case FMT_MOUNT_FULL:
        {
          sval = dp.name;
          break;
        }

        case FMT_SPECIAL:
        case FMT_SPECIAL_FULL:
        {
          sval = dp.special;
          break;
        }

        case FMT_TYPE:
        case FMT_TYPE_FULL:
        {
          sval = dp.fsType;
          break;
        }

        case FMT_BLOCKS_TOT:
        {
          rval = dp.totalBlocks;
          break;
        }

        case FMT_BLOCKS_TOT_AVAIL:
        {
          rval = dp.totalBlocks - (dp.freeBlocks - dp.availBlocks);
          break;
        }

        case FMT_BLOCKS_USED:
        {
          rval = dp.totalBlocks - dp.freeBlocks;
          break;
        }

        case FMT_BLOCKS_CALC_USED:
        {
          rval = dp.totalBlocks - dp.availBlocks;
          break;
        }

        case FMT_BLOCKS_FREE:
        {
          rval = dp.freeBlocks;
          break;
        }

        case FMT_BLOCKS_AVAIL:
        {
          rval = dp.availBlocks;
          break;
        }

        case FMT_BLOCKS_PERC_NOT_AVAIL:
        {
          rval = dp.totalBlocks;
          uval = dp.totalBlocks - dp.availBlocks;
          break;
        }

        case FMT_BLOCKS_PERC_USED:
        {
          rval = dp.totalBlocks;
          uval = dp.totalBlocks - dp.freeBlocks;
          break;
        }

        case FMT_BLOCKS_PERC_USED_AVAIL:
        {
          rval = dp.totalBlocks - (dp.freeBlocks - dp.availBlocks);
          uval = dp.totalBlocks - dp.freeBlocks;
          break;
        }

        case FMT_BLOCKS_PERC_AVAIL:
        {
          rval = dp.totalBlocks;
          uval = dp.availBlocks;
          break;
        }

        case FMT_BLOCKS_PERC_FREE:
        {
          rval = dp.totalBlocks;
          uval = dp.freeBlocks;
          break;
        }

        case FMT_INODES_TOT:
        {
          rval = dp.totalInodes;
          break;
        }

        case FMT_INODES_USED:
        {
          rval = dp.totalInodes - dp.freeInodes;
          break;
        }

        case FMT_INODES_FREE:
        {
          rval = dp.freeInodes;
          break;
        }

        case FMT_INODES_PERC:
        {
          rval = dp.totalInodes;
          uval = dp.totalInodes - dp.availInodes;
          break;
        }

        case FMT_MOUNT_TIME:
        {
          sval = dp.mountTime;
          break;
        }

        case FMT_MOUNT_OPTIONS:
        {
          sval = dp.mountOptions;
          break;
        }

        default:
        {
          break;
        }
      }  // switch on format character

      switch (ftype)
      {
        case FTYPE_STRING:
        {
          if (dispData.opts.csvOutput && ! first) {
            outString ~= ",";
          }
          outString ~= format (dispData.dispFmtString[c], sval);
          break;
        }

        case FTYPE_SPACE:
        {
          if (dispData.opts.csvOutput && ! first) {
            outString ~= ",";
          }
          outString ~= blockDisplay (dispData, c, rval, dHidx,
                dispData.opts.csvOutput);
          break;
        }

        case FTYPE_PERC_SPACE:
        case FTYPE_PERC_INODE:
        {
          if (dispData.opts.csvOutput && ! first) {
            outString ~= ",";
          }
          outString ~= percDisplay (dispData, c, rval, uval);
          break;
        }

        case FTYPE_INODE:
        {
          if (dispData.opts.csvOutput && ! first) {
            outString ~= ",";
          }
          outString ~= inodeDisplay (dispData, c, rval);
          break;
        }

        default:
        {
          break;
        }
      }
    } else {
      outString ~= c;
    }
    first = false;
  } // for each format character

  writeln (outString);
}

auto
findDispSize (real val)
{
  foreach (i, dt; displayTable)
  {
    if (val >= dt.low && val < dt.high)
    {
      return i;
    }
  }

  return displayIdxs['m'];
}

string
blockDisplay (const DisplayData dispData, const dchar c,
        const real val, const size_t dHidx, bool csvOutput)
{
  size_t        idx;
  real          dbs = dispData.dispOpts.dispBlockSize;
  string        suffix = "";

  if (dispData.dispOpts.dbsstr == "h") {
    idx = findDispSize (val);
  }
  if (dispData.dispOpts.dbsstr == "H") {
    idx = dHidx;
  }

  if (dispData.dispOpts.dbsstr == "h" || dispData.dispOpts.dbsstr == "H")
  {
    dbs = displayTable[idx].size;
    suffix = displayTable[idx].suffix;
  }

  auto nval = val / dbs;
  auto fmt = dispData.dispFmtString[c] ~ suffix;
  if (! csvOutput) {
    fmt ~= " ";
  } else {
    fmt = "\"" ~ fmt ~ "\"";
  }
  return format (fmt, nval);
}

string
inodeDisplay (const DisplayData dispData, const dchar c, const real val)
{
  size_t        idx;
  string        suffix = "";

  auto fmt = dispData.dispFmtString[c];
  return format (fmt, val);
}

string
percDisplay (const DisplayData dispData, const dchar c,
    const real tval, const real uval)
{
  real          perc;

  if (tval > 0.0)
  {
    perc = uval / tval;
    perc *= 100;
  }
  else
  {
    perc = 0.0;
  }

  auto fmt = dispData.dispFmtString[c];
  return format (fmt, perc);
}

void
setDispBlockSize (ref DisplayOpts dispOpts)
{
  real          val;
  char          c;

  val = 1024.0 * 1024.0;

  c = dispOpts.dbsstr[0];
  if (isDigit (c)) {
    val = to!(typeof(val))(dispOpts.dbsstr);
    val = floor (val);  // make sure it's an integer value
//writefln ("digit:%s:val=%25.0f", dispOpts.dbsstr, val);
    foreach (disp; displayTable) {
      if (val == disp.size) {
        c = disp.key;
//writefln ("digit:%s", c);
        break;
      }
    }
  } else {
    if (c != 'H') {
      c = cast(char) toLower (c);
    }
    if (dispOpts.dbsstr == "HUMAN") {
      c = 'h';
    }
//writefln ("non-digit:%s:%s", dispOpts.dbsstr, c);
    if (dispOpts.dbsstr.length > 1 && dispOpts.dbsstr[1] == 'B' &&
        dispOpts.baseDispSize != size1000) {
//writefln ("non-digit:1000:%s", c);
      dispOpts.baseDispSize = size1000;
      dispOpts.baseDispIdx = idx1000;
      initializeDisplayTable (dispOpts);
    }
    if (dispOpts.dbsstr.length > 1 && dispOpts.dbsstr[1] == 'b' &&
        dispOpts.baseDispSize != size1024) {
//writefln ("non-digit:1024:%s", c);
      dispOpts.baseDispSize = size1024;
      dispOpts.baseDispIdx = idx1024;
      initializeDisplayTable (dispOpts);
    }
  }

  if (c in displayIdxs) {
    auto idx = displayIdxs [c];
    dispOpts.dispBlockLabel = displayTable [idx].disp [dispOpts.baseDispIdx];
    val = displayTable [idx].size;
    dispOpts.precision = displayTable [idx].precision;
  } else {
    foreach (idx, disp; displayTable)
    {
      if (disp.key == 'h') { break; }
      auto tval = val / disp.size;
      if (floor (tval) == tval)
      {
        auto s = disp.suffix;
        if (idx == 0) { s = "b"; }
        dispOpts.dispBlockLabel = format ("%.0f%s", tval, s);
      }
    }
  }

  if (dispOpts.posixCompat && val == 512.0)
  {
    dispOpts.dispBlockLabel = "512-blocks";
  }
  if (dispOpts.posixCompat && val == 1024.0)
  {
    dispOpts.dispBlockLabel = "1024-blocks";
  }

//writefln ("block label:%s", dispOpts.dispBlockLabel);

  dispOpts.dispBlockSize = val;
//writefln ("block size:%.0f", val);
}

unittest {
  DisplayOpts       dispOpts;
  int               tcount;
  int               failures;

  void
  test (string l, bool posix,
      string dbsstr, real bds, int bdi,
      string rdbl, real rdbs)
  {
    bool      fail;

    ++tcount;
    dispOpts = dispOpts.init;
    dispOpts.dbsstr = dbsstr;
    dispOpts.baseDispSize = bds;
    dispOpts.baseDispIdx = bdi;
    dispOpts.posixCompat = posix;
    setDispBlockSize (dispOpts);
    if (dispOpts.dispBlockLabel != rdbl) { fail = true; }
    if (dispOpts.dispBlockSize != rdbs) { fail = true; }
    if (fail)
    {
      ++failures;
      writefln ("# %s: %s: %s",
        "setDispBlockSize:", "fail", l);
      writefln ("  expected: %s got %s", rdbl, dispOpts.dispBlockLabel);
      writefln ("  expected: %.0f got %.0f", rdbs, dispOpts.dispBlockSize);
    }
  }

  initializeIdxs ();

  dispOpts.baseDispSize = size1024;
  dispOpts.baseDispIdx = idx1024;
  initializeDisplayTable (dispOpts);
  for (auto i = 0; i < displayTable.length; ++i)
  {
    test (format("%.0f", displayTable[i].size), false,
          format ("%.0f", displayTable[i].size), size1024, idx1024,
          displayTable[i].disp[1], displayTable[i].size);
    test (to!string(displayTable[i].key), false,
          to!string(displayTable[i].key), size1024, idx1024,
          displayTable[i].disp[1], displayTable[i].size);
  }
  test ("HUMAN", false, "HUMAN", size1024, idx1024, "Size", 0);

  dispOpts.baseDispSize = size1000;
  dispOpts.baseDispIdx = idx1000;
  initializeDisplayTable (dispOpts);
  for (auto i = 0; i < displayTable.length; ++i)
  {
    test (format("%.0f", displayTable[i].size), false,
          format ("%.0f", displayTable[i].size), size1000, idx1000,
          displayTable[i].disp[0], displayTable[i].size);
    test (to!string(displayTable[i].key), false,
          to!string(displayTable[i].key), size1000, idx1000,
          displayTable[i].disp[0], displayTable[i].size);
  }
  test ("HUMAN", false, "HUMAN", size1000, idx1000, "Size", 0);

  dispOpts.baseDispSize = size1000;
  dispOpts.baseDispIdx = idx1000;
  initializeDisplayTable (dispOpts);
  for (auto i = 0; i < displayTable.length; ++i)
  {
    test (format ("%sB", displayTable[i].key), false,
          format ("%sB", displayTable[i].key), size1024, idx1024,
          displayTable[i].disp[0], displayTable[i].size);
  }

  test ("2.5", false, "2.5", size1000, idx1000, "2b", 2);
  test ("2", false, "2", size1000, idx1000, "2b", 2);
  test ("2048", false, "2048", size1000, idx1000, "2048b", 2048);
  test ("2000", false, "2000", size1000, idx1000, "2K", 2000);

  dispOpts.baseDispSize = size1024;
  dispOpts.baseDispIdx = idx1024;
  initializeDisplayTable (dispOpts);
  for (auto i = 0; i < displayTable.length; ++i)
  {
    test (format ("%sb", displayTable[i].key), false,
          format ("%sb", displayTable[i].key), size1000, idx1000,
          displayTable[i].disp[1], displayTable[i].size);
  }

  test ("2.5", false, "2.5", size1024, idx1024, "2b", 2.0);
  test ("2", false, "2", size1024, idx1024, "2b", 2.0);
  test ("2048", false, "2048", size1024, idx1024, "2K", 2048.0);
  test ("2000", false, "2000", size1024, idx1024, "2000b", 2000.0);

  test ("Posix-1024", true, "k", size1024, idx1024, "1024-blocks", 1024.0);
  test ("Posix-512", true, "512", size1024, idx1024, "512-blocks", 512.0);

  if (failures > 0) {
    write ("unittest: display: setDispBlockSize: ");
    writefln ("failed: %d of %d", failures, tcount);
  }
}

void
initializeDisplayTable (ref DisplayOpts dispOpts)
{
  displayTable [0].size = 1;
  displayTable [0].low = 1;     // temporary, reset below
  displayTable [0].high = dispOpts.baseDispSize;
  for (auto i = 1; i < displayTable.length; ++i)
  {
    if (displayTable [i].key == 'h') { break; }
    if (displayTable [i].key == 'H') { break; }
    displayTable [i].size = displayTable [i - 1].size * dispOpts.baseDispSize;
    displayTable [i].low = displayTable [i - 1].low * dispOpts.baseDispSize;
    displayTable [i].high = displayTable [i - 1].high * dispOpts.baseDispSize;
  }
  displayTable [0].low = 0;
}

void
initializeIdxs ()
{
  if (! dispIdxsInitialized) {
    foreach (idx, dt; displayTable) {
      displayIdxs[dt.key] = idx;
    }
    dispIdxsInitialized = true;
  }
  if (! fmtInfoIdxsInitialized) {
    foreach (idx, fi; formatTypes) {
      formatTypesIdxs[fi.key] = idx;
    }
    fmtInfoIdxsInitialized = true;
  }
}

void
setMaxLen (string val, ref int max)
{
  int len = cast(int) val.length;
  if (len > max)
  {
    max = len;
  }
}

void
initializeTitles (DisplayOpts dispOpts)
{
  if (dispOpts.posixCompat)
  {
    formatTypes[formatTypesIdxs[FMT_MOUNT_FULL]].title = "Mounted On";
    formatTypes[formatTypesIdxs[FMT_BLOCKS_AVAIL]].title = "Available";
    formatTypes[formatTypesIdxs[FMT_BLOCKS_PERC_NOT_AVAIL]].title = "Capacity";
    formatTypes[formatTypesIdxs[FMT_BLOCKS_PERC_USED]].title = "Capacity";
    formatTypes[formatTypesIdxs[FMT_BLOCKS_PERC_USED_AVAIL]].title = "Capacity";
  }
}

int
dpCompare (ref DiskPartition[] dps, string sortType, size_t i, size_t j)
{
  int       rc;
  int       sortOrder;

      /* reset sort order to the default start value */
  sortOrder = DI_SORT_ASCENDING;
  rc = 0;

  foreach (char st; sortType) {
    switch (st) {
      case DI_SORT_NONE: {
        break;
      }

      case DI_SORT_MOUNT: {
        rc = strcoll (toStringz(dps[i].name), toStringz(dps[j].name));
        rc *= sortOrder;
        break;
      }

      case DI_SORT_REVERSE: {
        sortOrder *= -1;
        break;
      }

      case DI_SORT_SPECIAL: {
        rc = strcoll (toStringz(dps[i].special), toStringz(dps[j].special));
        rc *= sortOrder;
        break;
      }

      case DI_SORT_TYPE: {
        rc = strcoll (toStringz(dps[i].fsType), toStringz(dps[j].fsType));
        rc *= sortOrder;
        break;
      }

      case DI_SORT_AVAIL: {
        auto v = dps[i].availBlocks - dps[j].availBlocks;
        rc = v < 0.00001 ? 0 : v > 0.0 ? 1 : -1;
        rc *= sortOrder;
        break;
      }

      default: {
        break;
      }
    } /* switch on sort type */

    if (rc != 0) {
      return rc;
    }
  }

  return rc;
}
